@@ -27,9 +27,7 @@
 
 #include <pthread.h>
 
-static volatile bool exiting = 0;
-
-struct stalld_bpf *stalld_obj;
+static struct stalld_bpf *stalld_obj;
 
 /*
  * Older versions of BPF does not have bpf_map__set_max_entries.
@@ -118,58 +116,6 @@ static int set_cpu_data(struct stalld_cpu_data *stalld_data, int cpu)
 	return 0;
 }
 
-/**
- * load_ebpf_context - sets up ebpf context
- *
- * Set up the basics for the ebpf program to run, raising
- * memlock limit, loading and attaching the eBPF code, set
- * up the perf buffer and return the ebpf object.
- */
-static struct stalld_bpf *load_ebpf_context(void)
-{
-	struct stalld_bpf *stalld_obj;
-	int err;
-
-	libbpf_set_print(libbpf_print_fn);
-
-	err = bump_memlock_rlimit();
-	if (err) {
-		warn("failed to increase rlimit: %d\n", err);
-		return 0;
-	}
-
-	stalld_obj = stalld_bpf__open();
-	if (!stalld_obj) {
-		warn("failed to open and/or load BPF object\n");
-		return 0;
-	}
-
-	err = bpf_map__set_max_entries(stalld_obj->maps.stalld_per_cpu_data, config_nr_cpus);
-	if (err) {
-		warn("failed to resize BPF map: %d\n", err);
-		return 0;
-	} else {
-		log_msg("adjusted stalld map to %d cpus\n", config_nr_cpus);
-	}
-
-	err = stalld_bpf__load(stalld_obj);
-	if (err) {
-		warn("failed to load BPF object: %d\n", err);
-		goto cleanup;
-	}
-
-	err = stalld_bpf__attach(stalld_obj);
-	if (err) {
-		warn("failed to attach BPF programs\n");
-		goto cleanup;
-	}
-
-	return stalld_obj;
-
-cleanup:
-	stalld_bpf__destroy(stalld_obj);
-	return 0;
-}
 
 static int queue_track_get_cpu(char *buffer, int size, int cpu)
 {
@@ -201,11 +147,7 @@ static int queue_track_parse(struct cpu_info *cpu_info, char *buffer, size_t buf
 	struct queued_task *qtask;
 	int retval = 0;
 
-	tasks = calloc(MAX_QUEUE_TASK, sizeof(struct task_info));
-	if (tasks == NULL) {
-		warn("failed to malloc %d task_info structs", MAX_QUEUE_TASK);
-		goto error;
-	}
+	tasks = allocate_memory(MAX_QUEUE_TASK, sizeof(struct task_info));
 
 	for_each_queued_task(cpu_data, qtask) {
 		if (qtask->is_rt)
@@ -253,9 +195,6 @@ static int queue_track_parse(struct cpu_info *cpu_info, char *buffer, size_t buf
         }
 
 	return 0;
-
-error:
-	return 1;
 }
 
 static int queue_track_has_starving_task(struct cpu_info *cpu)
@@ -263,20 +202,23 @@ static int queue_track_has_starving_task(struct cpu_info *cpu)
 	return !!cpu->nr_rt_running;
 }
 
-static int queue_track_init(void)
+/**
+ * initialize_maps - Initialize BPF per-CPU data maps
+ *
+ * This function initializes the BPF maps used for per-CPU monitoring data.
+ * It retrieves existing CPU data from the BPF map, enables monitoring for
+ * configured CPUs, and updates the map with the new monitoring state.
+ *
+ * Returns: 0 on success, -1 on error
+ */
+static int initialize_maps(void)
 {
 	struct stalld_cpu_data stalld_data;
-	int retval, i;
 
-	stalld_obj = load_ebpf_context();
-	if (!stalld_obj)
-		return 1;
-
-	for (i = 0; i < config_nr_cpus; i++) {
+	for (int i = 0; i < config_nr_cpus; i++) {
 		/* Init data */
-		retval = get_cpu_data(&stalld_data, i);
-		if (retval)
-			goto destroy;
+		if (get_cpu_data(&stalld_data, i))
+			return -1;
 
 		if (config_monitor_all_cpus || config_monitored_cpus[i])
 			stalld_data.monitoring = 1;
@@ -287,10 +229,127 @@ static int queue_track_init(void)
 	/* it is static */
 	config_buffer_size = sizeof(struct stalld_cpu_data);
 	return 0;
+}
+
+/**
+ * run_task_iterator - Execute the BPF task iterator
+ *
+ * This function creates and runs the BPF task iterator program to walk
+ * through all tasks in the system. The iterator provides a snapshot view
+ * of all tasks, complementing the event-driven tracepoint monitoring.
+ *
+ * Returns: 0 on success, negative value on error
+ */
+static int run_task_iterator(void)
+{
+	struct bpf_link *iter_link;
+	char buf[64];
+	int iter_fd, len;
+
+	if (!stalld_obj) {
+		warn("BPF object not loaded\n");
+		return -EINVAL;
+	}
+
+	/* Create the iterator link */
+	iter_link = bpf_program__attach_iter(stalld_obj->progs.iter_task, NULL);
+	if (!iter_link) {
+		warn("Failed to attach task iterator\n");
+		return -EINVAL;
+	}
+
+	/* Get file descriptor for the iterator */
+	iter_fd = bpf_iter_create(bpf_link__fd(iter_link));
+	if (iter_fd < 0) {
+		warn("Failed to create iterator fd: %d\n", iter_fd);
+		bpf_link__destroy(iter_link);
+		return iter_fd;
+	}
+
+	/* Run the iterator - this will trigger iteration through all tasks */
+	while ((len = read(iter_fd, buf, sizeof(buf))) > 0) {
+		/* Iterator output is processed by the BPF program itself */
+		/* The actual task tracking happens in the BPF program */
+	}
+
+	if (len < 0)
+		warn("Iterator read error: %d\n", len);
+
+	close(iter_fd);
+	bpf_link__destroy(iter_link);
+
+	log_verbose("Task iterator completed\n");
+	return len < 0 ? len : 0;
+}
+
+/**
+ * load_ebpf_context - sets up ebpf context
+ *
+ * Set up the basics for the ebpf program to run, raising
+ * memlock limit, loading and attaching the eBPF code, set
+ * up the perf buffer and return the ebpf object.
+ */
+static int load_ebpf_context(void)
+{
+	int err;
+
+	libbpf_set_print(libbpf_print_fn);
+
+	err = bump_memlock_rlimit();
+	if (err) {
+		warn("failed to increase rlimit: %d\n", err);
+		return -1;
+	}
+
+	stalld_obj = stalld_bpf__open();
+	if (!stalld_obj) {
+		warn("failed to open and/or load BPF object\n");
+		return -1;
+	}
+
+	err = bpf_map__set_max_entries(stalld_obj->maps.stalld_per_cpu_data, config_nr_cpus);
+	if (err) {
+		warn("failed to resize BPF map: %d\n", err);
+		goto cleanup;
+	} else {
+		log_msg("adjusted stalld map to %d cpus\n", config_nr_cpus);
+	}
+
+
+	err = stalld_bpf__load(stalld_obj);
+	if (err) {
+		warn("failed to load BPF object: %d\n", err);
+		goto cleanup;
+	}
+
+	return 0;
+
+cleanup:
+	stalld_bpf__destroy(stalld_obj);
+	return -1;
+}
+
+static int queue_track_init(void)
+{
+	if (load_ebpf_context())
+		return -1;
+
+	if (initialize_maps())
+		goto destroy;
+
+	if (run_task_iterator())
+		goto destroy;
+
+	if (stalld_bpf__attach(stalld_obj)) {
+		warn("failed to attach BPF programs\n");
+		goto destroy;
+	}
+
+	return 0;
 
 destroy:
 	stalld_bpf__destroy(stalld_obj);
-	return 1;
+	return -1;
 }
 
 static void queue_track_destroy(void)
