@@ -15,13 +15,11 @@
  */
 
 #define _GNU_SOURCE
-#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
 #include <pthread.h>
 #include <sched.h>
-#include <signal.h>
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -38,9 +36,6 @@
 
 #include "stalld.h"
 #include "sched_debug.h"
-#if USE_BPF
-#include "queue_track.h"
-#endif
 
 /*
  * version
@@ -165,95 +160,6 @@ struct stalld_backend *backend = &sched_debug_backend;
  */
 char *config_affinity_cpus;
 
-/*
- * API to fetch the process group ID for a thread/process.
- */
-int get_tgid(int pid)
-{
-	const char tgid_field[TGID_FIELD] = "Tgid:";
-	char file_location[PROC_PID_FILE_PATH_LEN];
-	char *status = NULL;
-	int tgid, n;
-	FILE *fp;
-
-	status = calloc(TMP_BUFFER_SIZE, sizeof(char));
-	if (status == NULL) {
-		return -ENOMEM;
-	}
-
-	n = sprintf(file_location, "/proc/%d/status", pid);
-	if (n < 0)
-		goto out_free_mem;
-
-	if ((fp = fopen(file_location, "r")) == NULL)
-		goto out_free_mem;
-
-	/* Iterate till we find the tgid field. */
-	while (1) {
-		if (fgets(status, TMP_BUFFER_SIZE, fp) == NULL)
-			goto out_close_fd;
-		if (!(strncmp(status, tgid_field, (TGID_FIELD - 1))))
-			break;
-		/*
-		 * Zero out the buffer just in case
-		 */
-		memset(status, 0, TMP_BUFFER_SIZE);
-	}
-	/*
-	 * Since we're now at the line we're interested in, let's read
-	 * in the field that we want.
-	 */
-	if (sscanf(status, "%*s %d", &tgid) != 1)
-		goto out_close_fd;
-
-	fclose(fp);
-	free(status);
-	return tgid;
-
-out_close_fd:
-	fclose(fp);
-out_free_mem:
-	free(status);
-	return -EINVAL;
-}
-
-/*
- * Read the content of /proc/stat into the input buffer.
- * Used by functions doing cpu idle detection
- */
-int read_proc_stat(char *buffer, int size)
-{
-	int position = 0;
-	int retval;
-	int fd;
-
-	fd = open("/proc/stat", O_RDONLY);
-
-	if (fd < 0)
-		goto out_error;
-
-	do {
-		retval = read(fd, &buffer[position], size - position);
-		if (retval < 0)
-			goto out_close_fd;
-
-		position += retval;
-
-	} while (retval > 0 && position < size);
-
-	buffer[position-1] = '\0';
-
-	close(fd);
-
-	return position;
-
-out_close_fd:
-	close(fd);
-
-out_error:
-	return 0;
-}
-
 /*
  * calculate a buffer size to use when reading /proc/stat
  */
@@ -291,57 +197,22 @@ static int calc_stat_max(int pgsize)
  */
 static long get_cpu_idle_time(char *buffer, size_t buffer_size, int cpu)
 {
-	char cpuid[10]; /* cpuXXXXX\n */
-	char *idle_start;
-	char *end;
-	long val;
+	char cpuid[10];
+	char *line;
+	long user, nice, system, idle;
+	int ret;
 
 	sprintf(cpuid, "cpu%d ", cpu);
 
-	/* CPU */
-	idle_start = strstr(buffer, cpuid);
-	if (!idle_start)
+	line = strstr(buffer, cpuid);
+	if (!line)
 		return -ENODEV; /* CPU might be offline. */
 
-	/* Find and skip space before user. */
-	idle_start = strstr(idle_start, " ");
-	if (!idle_start)
+	ret = sscanf(line, "cpu%d %ld %ld %ld %ld", &cpu, &user, &nice, &system, &idle);
+	if (ret != 5)
 		return -EINVAL;
 
-	idle_start+=1;
-
-	/* Find and skip space before nice. */
-	idle_start = strstr(idle_start, " ");
-	if (!idle_start)
-		return -EINVAL;
-
-	idle_start+=1;
-
-	/* Find and skip space before system. */
-	idle_start = strstr(idle_start, " ");
-	if (!idle_start)
-		return -EINVAL;
-
-	idle_start+=1;
-
-	/* Here is the idle! */
-	idle_start = strstr(idle_start, " ");
-	if (!idle_start)
-		return -EINVAL;
-
-	idle_start += 1;
-
-	/* End. */
-	end = strstr(idle_start, " ");
-	if (!end)
-		return -EINVAL;
-
-	errno = 0;
-	val = strtol(idle_start, &end, 10);
-	if (errno != 0)
-		return -EINVAL;
-
-	return val;
+	return idle;
 }
 
 int cpu_had_idle_time(struct cpu_info *cpu_info)
@@ -462,7 +333,7 @@ struct cpu_starving_task_info {
 
 struct cpu_starving_task_info *cpu_starving_vector;
 
-void update_cpu_starving_vector(int cpu, int tgid, int pid, time_t since, struct task_info *task)
+void update_cpu_starving_vector(int cpu, const struct task_info *task)
 {
 	struct cpu_starving_task_info *cpu_info = &cpu_starving_vector[cpu];
 
@@ -475,23 +346,34 @@ void update_cpu_starving_vector(int cpu, int tgid, int pid, time_t since, struct
 
 	/*
 	 * If there is no thread in the vector, or if the in the
-	 * vector has an earlier since (time stamp), update it.
+	 * vector has an earlier since (timestamp), update it.
+	 * Also, check if the PIDs match if detect task migration.
 	 */
-	if ((cpu_info->since == 0) || cpu_info->since > since) {
+	if (cpu_info->since == 0
+	    || cpu_info->since > task->since
+	    || cpu_info->pid != task->pid) {
 		memcpy(&(cpu_info->task), task, sizeof(struct task_info));
-		cpu_info->pid = pid;
-		cpu_info->tgid = tgid;
-		cpu_info->since = since;
+		cpu_info->pid = task->pid;
+		cpu_info->tgid = task->tgid;
+		cpu_info->since = task->since;
 	}
 }
 
 void merge_taks_info(int cpu, struct task_info *old_tasks, int nr_old, struct task_info *new_tasks, int nr_new)
 {
+	static const struct task_info notask = {};
+
 	struct task_info *old_task;
 	struct task_info *new_task;
 	int i;
 	int j;
 
+	if (nr_new == 0) {
+		/* no starving tasks, reset the starving vector */
+		update_cpu_starving_vector(cpu, &notask);
+		return;
+	}
+
 	for (i = 0; i < nr_old; i++) {
 		old_task = &old_tasks[i];
 
@@ -502,7 +384,7 @@ void merge_taks_info(int cpu, struct task_info *old_tasks, int nr_old, struct ta
 				if (old_task->ctxsw == new_task->ctxsw) {
 					new_task->since = old_task->since;
 					if (config_single_threaded)
-						update_cpu_starving_vector(cpu, new_task->tgid, new_task->pid, new_task->since, new_task);
+						update_cpu_starving_vector(cpu, new_task);
 				}
 				break;
 			}
@@ -510,6 +392,23 @@ void merge_taks_info(int cpu, struct task_info *old_tasks, int nr_old, struct ta
 	}
 }
 
+/**
+ * cleanup_starving_task_info - Reset a CPU's starving task info structure
+ * @info: Pointer to the cpu_starving_task_info structure to clean up
+ *
+ * Clears all fields in the starving task info structure, effectively removing
+ * any tracked starving task for this CPU. This function preserves the overloaded
+ * flag value before clearing the structure and returns it to the caller.
+ *
+ * Return: The previous value of the overloaded flag before cleanup
+ */
+static int cleanup_starving_task_info(struct cpu_starving_task_info *info)
+{
+	const int overloaded = info->overloaded;
+	bzero(info, sizeof *info);
+	return overloaded;
+}
+
 int get_current_policy(int pid, struct sched_attr *attr)
 {
 	int ret;
@@ -720,35 +619,36 @@ int check_starving_tasks(struct cpu_info *cpu)
 	for (i = 0; i < cpu->nr_waiting_tasks; i++) {
 		task = &tasks[i];
 
-		if ((time(NULL) - task->since) >= config_starving_threshold) {
+		/* Skip tasks that haven't been starving long enough */
+		if ((time(NULL) - task->since) < config_starving_threshold)
+			continue;
 
-			log_msg("%s-%d starved on CPU %d for %d seconds\n",
-				task->comm, task->pid, cpu->id,
-				(time(NULL) - task->since));
+		log_msg("%s-%d starved on CPU %d for %d seconds\n",
+			task->comm, task->pid, cpu->id,
+			(time(NULL) - task->since));
 
-			/*
-			 * Check if this task needs to be ignored from being boosted
-			 * if yes, update the time stamp so that it doesn't keep
-			 * getting reported as being starved.
-			 */
-			if (config_ignore && !(check_task_ignore(task))) {
-				task->since = time(NULL);
-				continue;
-			}
-
-			starving+=1;
-
-			/*
-			 * It it is only logging, just reset the time counter
-			 * after logging.
-			 */
-			if (config_log_only) {
-				task->since = time(NULL);
-				continue;
-			}
-
-			boost_starving_task(task->tgid, task->pid, cpu);
+		/*
+		 * Check if this task needs to be ignored from being boosted
+		 * if yes, update the time stamp so that it doesn't keep
+		 * getting reported as being starved.
+		 */
+		if (config_ignore && !(check_task_ignore(task))) {
+			task->since = time(NULL);
+			continue;
 		}
+
+		starving++;
+
+		/*
+		 * If it is only logging, just reset the time counter
+		 * after logging.
+		 */
+		if (config_log_only) {
+			task->since = time(NULL);
+			continue;
+		}
+
+		boost_starving_task(task->tgid, task->pid, cpu);
 	}
 
 	return starving;
@@ -828,16 +728,7 @@ void *cpu_main(void *data)
 	while (cpu->thread_running && running) {
 
 		/* Buffer size should increase. See sched_debug_get(). */
-		if (config_buffer_size != cpu->buffer_size) {
-			char *old_buffer = cpu->buffer;
-			cpu->buffer = realloc(cpu->buffer, config_buffer_size);
-			if (!cpu->buffer) {
-				warn("fail to increase the buffer... continue");
-				cpu->buffer = old_buffer;
-			} else {
-				cpu->buffer_size = config_buffer_size;
-			}
-		}
+		resize_buffer_if_needed(&cpu->buffer, &cpu->buffer_size);
 
 		if (config_idle_detection) {
 			if (cpu_had_idle_time(cpu)) {
@@ -885,6 +776,27 @@ static const char *join_thread(pthread_t *thread)
 	return result;
 }
 
+/*
+ * Check if idle detection should skip parsing.
+ * Returns 1 if parsing should be skipped, 0 otherwise.
+ */
+static int should_skip_idle_cpus(struct cpu_info *cpus, int nr_cpus, char *busy_cpu_list)
+{
+	int has_busy_cpu;
+
+	if (!config_idle_detection)
+		return 0;
+
+	memset(busy_cpu_list, 0, nr_cpus);
+	has_busy_cpu = get_cpu_busy_list(cpus, nr_cpus, busy_cpu_list);
+	if (!has_busy_cpu) {
+		log_verbose("all CPUs had idle time, skipping parse\n");
+		return 1;
+	}
+
+	return 0;
+}
+
 void aggressive_main(struct cpu_info *cpus, int nr_cpus)
 {
 	int i;
@@ -913,14 +825,10 @@ void conservative_main(struct cpu_info *cpus, int nr_cpus)
 	size_t buffer_size = 0;
 	struct cpu_info *cpu;
 	char *buffer = NULL;
-	int has_busy_cpu;
 	int retval;
 	int i;
 
-	buffer = malloc(config_buffer_size);
-	if (!buffer)
-		die("cannot allocate buffer");
-
+	buffer = allocate_memory(config_buffer_size, sizeof *buffer);
 	buffer_size = config_buffer_size;
 
 	pthread_attr_init(&dettached);
@@ -934,25 +842,10 @@ void conservative_main(struct cpu_info *cpus, int nr_cpus)
 	while (running) {
 
 		/* Buffer size should increase. See sched_debug_get(). */
-		if (config_buffer_size != buffer_size) {
-			char *old_buffer = buffer;
-			buffer = realloc(buffer, config_buffer_size);
-			if (!buffer) {
-				warn("fail to increase the buffer... continue");
-				buffer = old_buffer;
-			} else {
-				buffer_size = config_buffer_size;
-			}
-		}
+		resize_buffer_if_needed(&buffer, &buffer_size);
 
-		if (config_idle_detection) {
-			memset(&busy_cpu_list, 0, nr_cpus);
-			has_busy_cpu = get_cpu_busy_list(cpus, nr_cpus, busy_cpu_list);
-			if (!has_busy_cpu) {
-				log_verbose("all CPUs had idle time, skipping parse\n");
-				goto skipped;
-			}
-		}
+		if (should_skip_idle_cpus(cpus, nr_cpus, busy_cpu_list))
+			goto skipped;
 
 		if (backend->get) {
 			retval = backend->get(buffer, buffer_size);
@@ -1022,33 +915,28 @@ int boost_cpu_starving_vector(struct cpu_starving_task_info *vector, int nr_cpus
 		if (config_log_only)
 			continue;
 
-		if (cpu->pid != 0 && (now - cpu->since) > config_starving_threshold) {
-			/*
-			 * Check if this task name is part of a denylist
-			 * If yes, do not boost it.
-			 */
-			if (config_ignore && !check_task_ignore(&cpu->task))
-				continue;
+		/* Skip if no task or not starving long enough */
+		if (cpu->pid == 0 || (now - cpu->since) <= config_starving_threshold)
+			continue;
 
-			/* Save the task policy. */
-			ret = get_current_policy(cpu->pid, &attr[i]);
-
-			/* It is ok if a task die. */
-			if (ret < 0)
-				continue;
+		/* Skip if task is on denylist */
+		if (config_ignore && !check_task_ignore(&cpu->task))
+			continue;
 
+		/* Save the task policy. */
+		ret = get_current_policy(cpu->pid, &attr[i]);
+		if (!ret) /* It is ok if a task die. */
 			/* Boost! */
 			ret = boost_with_deadline(cpu->tgid, cpu->pid, &cpus[i]);
 
-			/* It is ok if a task die. */
-			if (ret < 0)
-				continue;
-
-			/* Save it for the deboost. */
-			deboost_vector[i] = cpu->pid;
-
-			boosted++;
+		if (ret < 0) {
+			cleanup_starving_task_info(cpu);
+			continue;
 		}
+
+		/* Save it for the deboost. */
+		deboost_vector[i] = cpu->pid;
+		boosted++;
 	}
 
 	if (!boosted)
@@ -1071,7 +959,6 @@ void single_threaded_main(struct cpu_info *cpus, int nr_cpus)
 	struct cpu_info *cpu;
 	char *buffer = NULL;
 	int overloaded = 0;
-	int has_busy_cpu;
 	int boosted = 0;
 	int retval;
 	int i;
@@ -1081,13 +968,8 @@ void single_threaded_main(struct cpu_info *cpus, int nr_cpus)
 	if (!config_log_only && boost_policy != SCHED_DEADLINE)
 		die("Single threaded mode only works with SCHED_DEADLINE");
 
-	cpu_starving_vector = malloc(sizeof(struct cpu_starving_task_info) * nr_cpus);
-	if (!cpu_starving_vector)
-		die("cannot allocate cpu starving vector");
-
-	buffer = malloc(config_buffer_size);
-	if (!buffer)
-		die("cannot allocate buffer");
+	cpu_starving_vector = allocate_memory(nr_cpus, sizeof(struct cpu_starving_task_info));
+	buffer = allocate_memory(config_buffer_size, sizeof *buffer);
 
 	buffer_size = config_buffer_size;
 
@@ -1103,25 +985,10 @@ void single_threaded_main(struct cpu_info *cpus, int nr_cpus)
 	while (running) {
 
 		/* Buffer size should increase. See sched_debug_get(). */
-		if (config_buffer_size != buffer_size) {
-			char *old_buffer = buffer;
-			buffer = realloc(buffer, config_buffer_size);
-			if (!buffer) {
-				warn("fail to increase the buffer... continue");
-				buffer = old_buffer;
-			} else {
-				buffer_size = config_buffer_size;
-			}
-		}
+		resize_buffer_if_needed(&buffer, &buffer_size);
 
-		if (config_idle_detection) {
-			memset(&busy_cpu_list, 0, nr_cpus);
-			has_busy_cpu = get_cpu_busy_list(cpus, nr_cpus, busy_cpu_list);
-			if (!has_busy_cpu) {
-				log_verbose("all CPUs had idle time, skipping parse\n");
-				goto skipped;
-			}
-		}
+		if (should_skip_idle_cpus(cpus, nr_cpus, busy_cpu_list))
+			goto skipped;
 
 		if (backend->get) {
 			retval = backend->get(buffer, buffer_size);
@@ -1155,12 +1022,8 @@ void single_threaded_main(struct cpu_info *cpus, int nr_cpus)
 
 		/* Cleanup the CPU starving vector. */
 		for (i = 0; i < nr_cpus; i++) {
-			memset(&(cpu_starving_vector[i].task), 0, sizeof(struct task_info));
-			cpu_starving_vector[i].pid = 0;
-			cpu_starving_vector[i].since = 0;
-			if (cpu_starving_vector[i].overloaded)
+			if (cleanup_starving_task_info(cpu_starving_vector+i))
 				overloaded = 1;
-			cpu_starving_vector[i].overloaded = 0;
 		}
 
 		/*
@@ -1299,17 +1162,10 @@ int main(int argc, char **argv)
 	if (!config_log_only)
 		boost_policy = check_policies();
 
-	cpus = malloc(sizeof(struct cpu_info) * config_nr_cpus);
-	if (!cpus)
-		die("Cannot allocate memory");
-
-	memset(cpus, 0, sizeof(struct cpu_info) * config_nr_cpus);
+	cpus = allocate_memory(config_nr_cpus, sizeof(struct cpu_info));
 
 	for (i = 0; i < config_nr_cpus; i++) {
-		cpus[i].buffer = malloc(config_buffer_size);
-		if (!cpus[i].buffer)
-			die("Cannot allocate memory");
-
+		cpus[i].buffer = allocate_memory(1, config_buffer_size);
 		cpus[i].buffer_size = config_buffer_size;
 	}
 
