@@ -28,10 +28,48 @@
 #include <sched.h>
 #include <linux/sched.h>
 #include <sys/sysinfo.h>
+#include <mntent.h>
 
 #include "stalld.h"
 #include "sched_debug.h"
+#if USE_BPF
 #include "queue_track.h"
+#endif
+
+static int find_debugfs_mount_point(char *mount_path_buf, size_t buf_size);
+
+/*
+ * Helper function to resize buffer when needed.
+ * Returns 0 on success, -1 on failure.
+ */
+int resize_buffer_if_needed(char **buffer, size_t *current_size)
+{
+	if (config_buffer_size == *current_size)
+		return 0;
+
+	char *old_buffer = *buffer;
+	*buffer = realloc(*buffer, config_buffer_size);
+	if (!*buffer) {
+		warn("fail to increase the buffer... continue");
+		*buffer = old_buffer;
+		return -1;
+	}
+
+	*current_size = config_buffer_size;
+	return 0;
+}
+
+/*
+ * Helper function for memory allocation with error checking.
+ * Dies with an error message if allocation fails.
+ */
+void *allocate_memory(size_t count, size_t size)
+{
+	void *ptr = calloc(count, size);
+	if (!ptr)
+		die("Cannot allocate memory");
+	return ptr;
+}
 
 /*
  * fill_process_comm - process name from task group ID.
@@ -53,7 +91,7 @@ int fill_process_comm(int tgid, int pid, char *comm, int comm_size)
 	if (retval < 0)
 		goto out_error;
 
-	fd = open(path, O_RDWR);
+	fd = open(path, O_RDONLY);
 	if (fd < 0) {
 		log_msg("failed to open comm file at %s\n", path);
 		goto out_error;
@@ -213,14 +251,10 @@ int setup_signal_handling(void)
  */
 void __die(const char *fmt, ...)
 {
-	volatile int zero = 0;
-	int ret = errno;
 	va_list ap;
 
 	if (errno)
 		perror("stalld");
-	else
-		ret = -1;
 
 	va_start(ap, fmt);
 	fprintf(stderr, "  ");
@@ -229,11 +263,11 @@ void __die(const char *fmt, ...)
 
 	fprintf(stderr, "\n");
 
-	/* Die with a division by zero to keep the stack on GDB. */
+	/* in verbose mode, abort() to generate a crash dump */
 	if (config_verbose)
-		zero = 10 / zero;
+		abort();
 
-	exit(ret);
+	exit(EXIT_FAILURE);
 }
 
 /*
@@ -404,44 +438,6 @@ void cleanup_regex(unsigned int *nr_task, regex_t **compiled_expr)
 #define _STR(x) #x
 #define STR(x) _STR(x)
 
-static int find_mount(const char *mount, char *debugfs)
-{
-	char type[100];
-	FILE *fp;
-
-	if ((fp = fopen("/proc/mounts","r")) == NULL)
-		return 0;
-
-	while (fscanf(fp, "%*s %"
-		      STR(MAX_PATH)
-		      "s %99s %*s %*d %*d\n",
-		      debugfs, type) == 2) {
-		if (strcmp(type, mount) == 0)
-			break;
-	}
-	fclose(fp);
-
-	if (strcmp(type, mount) != 0)
-		return 0;
-	return 1;
-}
-
-static const char *find_debugfs(void)
-{
-	static char debugfs[MAX_DIR_PATH];
-	static int debugfs_found;
-
-	if (debugfs_found)
-		return debugfs;
-
-	if (!find_mount("debugfs", debugfs))
-		return "";
-
-	debugfs_found = 1;
-
-	return debugfs;
-}
-
 /*
  * Return true if the file at *path exists.
  */
@@ -462,19 +458,20 @@ static int check_file_exists(char *path)
  */
 static int find_debugfs_sched_debug(void)
 {
-	const char *debugfs;
+	char debugfs[MAX_DIR_PATH];
 	char *path;
 	int found;
+	size_t size;
 
-	debugfs = find_debugfs();
-	if (!debugfs)
+	if(find_debugfs_mount_point(debugfs, sizeof(debugfs)))
 		return 0;
 
-	path = malloc(strlen(debugfs) + strlen("sched/debug") + 1);
+	size = strlen(debugfs) + strlen("sched/debug") + 2;
+	path = malloc(size);
 	if (!path)
 		return 0;
 
-	sprintf(path, "%s/%s", debugfs, "sched/debug");
+	snprintf(path, size, "%s/%s", debugfs, "sched/debug");
 
 	found = check_file_exists(path);
 	if (found)
@@ -564,11 +561,10 @@ out_err:
 
 static int fill_sched_features_path(char *path, int path_size)
 {
-	const char *debugfs;
+	char debugfs[MAX_DIR_PATH];
 	int retval;
 
-	debugfs = find_debugfs();
-	if (strlen(debugfs) == 0)
+	if (find_debugfs_mount_point(debugfs, sizeof(debugfs)))
 		return 0;
 
 	snprintf(path, path_size, "%s/sched/features", debugfs);
@@ -782,7 +778,9 @@ static void print_usage(void)
 		"	backend:",
 		"	   -b/--backend: backend to be used to search for starving tasks, options are:",
 		"		sched_debug || S: for sched/debug file,",
+#if USE_BPF
 		"		queue_track || Q: for tracking enqueue/dequeue of tasks in the runqueues.",
+#endif
 		"	misc:",
 		"          --pidfile: write daemon pid to specified file",
 		"          -S/--systemd: running as systemd service, don't fiddle with RT throttling",
@@ -962,13 +960,11 @@ static void parse_cpu_list(char *cpulist)
 			end_cpu = cpu;
 
 		if (cpu == end_cpu) {
-			if (config_verbose)
-				printf("cpulist: adding cpu %d\n", cpu);
+			info("cpulist: adding cpu %d\n", cpu);
 			config_monitored_cpus[cpu] = 1;
 		} else {
 			for (i = cpu; i <= end_cpu; i++) {
-				if (config_verbose)
-					printf("cpulist: adding cpu %d\n", i);
+				info("cpulist: adding cpu %d\n", i);
 				config_monitored_cpus[i] = 1;
 			}
 		}
@@ -1151,9 +1147,11 @@ int parse_args(int argc, char **argv)
 			if (!strcmp(optarg, "sched_debug") || !strcmp(optarg, "S")) {
 				backend = &sched_debug_backend;
 				log_msg("using sched_debug backend\n");
+#if USE_BPF
 			} else if (!strcmp(optarg, "queue_track") || !strcmp(optarg, "Q")) {
 				backend = &queue_track_backend;
 				log_msg("using queue_track backend\n");
+#endif
 			} else {
 				usage("unknown backend %s\n", optarg);
 			}
@@ -1216,3 +1214,175 @@ int parse_args(int argc, char **argv)
 
 	return(0);
 }
+
+/**
+ * Dynamically finds the mount point of the debugfs filesystem.
+ *
+ * This function reads through /proc/mounts to find an entry with the
+ * filesystem type "debugfs".
+ *
+ * @param mount_path_buf A buffer to store the found debugfs mount path.
+ * @param buf_size The size of the mount_path_buf.
+ * @return 0 on success (debugfs mount point found), -1 if not found or on error.
+ */
+static int find_debugfs_mount_point(char *mount_path_buf, size_t buf_size) {
+	FILE *fp;
+	struct mntent *mnt;
+	int ret = 0;
+
+	fp = setmntent("/proc/mounts", "r");
+	if (!fp) {
+		warn("Error opening /proc/mounts");
+		return -1;
+	}
+
+	while ((mnt = getmntent(fp))) {
+		if (strcmp(mnt->mnt_type, "debugfs"))
+			continue;
+
+		if (strlen(mnt->mnt_dir) < buf_size) {
+			strncpy(mount_path_buf, mnt->mnt_dir, buf_size - 1);
+			mount_path_buf[buf_size - 1] = '\0';
+		} else {
+			warn("Buffer too small for debugfs mount path: %s\n", mnt->mnt_dir);
+			ret = -1;
+		}
+
+		break;
+	}
+
+	endmntent(fp);
+	return ret;
+}
+
+/**
+ * Checks if the 'sched/fair_server' directory exists within debugfs,
+ * dynamically determining the debugfs mount path.
+ *
+ * This function first attempts to find the mount point of the debugfs
+ * filesystem. If found, it then constructs the full path to the
+ * 'sched/fair_server' directory within that mount point. Finally, it uses
+ * the stat() system call to query information about this path and determines
+ * if it exists and is indeed a directory.
+ *
+ * @return 0 if the directory exists and is a directory.
+ * Returns -1 if debugfs is not mounted, or other system error occurs.
+ */
+int check_dl_server_dir_exists(void) {
+	char debugfs_mount_path[MAX_DIR_PATH];
+	char full_target_path[MAX_PATH];
+	struct stat st;
+	const char *dl_server_subdir = "/sched/fair_server";
+
+	if (find_debugfs_mount_point(debugfs_mount_path, sizeof(debugfs_mount_path))) {
+		warn("Debugfs mount point not found or error during lookup.\n");
+		return -1;
+	}
+
+	if (snprintf(full_target_path, sizeof(full_target_path), "%s%s",debugfs_mount_path,
+		     dl_server_subdir) >= sizeof(full_target_path)) {
+		errno = ENAMETOOLONG;
+		warn("Constructed path too long for buffer");
+		return -1;
+	}
+
+	if (stat(full_target_path, &st) == 0) {
+		if (S_ISDIR(st.st_mode))
+			return 0;
+
+		warn("Path '%s' exists but is not a directory.\n", full_target_path);
+		return -1;
+	}
+
+	log_msg("DL-server not detected: %s\n", strerror(errno));
+	return -1;
+}
+
+/*
+ * API to fetch the process group ID for a thread/process.
+ */
+int get_tgid(int pid)
+{
+	const char tgid_field[] = "Tgid:";
+	char file_location[PROC_PID_FILE_PATH_LEN];
+	char *status = NULL;
+	int tgid, n;
+	FILE *fp;
+
+	status = calloc(TMP_BUFFER_SIZE, sizeof(char));
+	if (status == NULL) {
+		return -ENOMEM;
+	}
+
+	n = sprintf(file_location, "/proc/%d/status", pid);
+	if (n < 0)
+		goto out_free_mem;
+
+	if ((fp = fopen(file_location, "r")) == NULL)
+		goto out_free_mem;
+
+	/* Iterate till we find the tgid field. */
+	while (1) {
+		if (fgets(status, TMP_BUFFER_SIZE, fp) == NULL)
+			goto out_close_fd;
+		if (!(strncmp(status, tgid_field, sizeof(tgid_field) - 1)))
+			break;
+		/*
+		 * Zero out the buffer just in case
+		 */
+		memset(status, 0, TMP_BUFFER_SIZE);
+	}
+	/*
+	 * Since we're now at the line we're interested in, let's read
+	 * in the field that we want.
+	 */
+	if (sscanf(status, "%*s %d", &tgid) != 1)
+		goto out_close_fd;
+
+	fclose(fp);
+	free(status);
+	return tgid;
+
+out_close_fd:
+	fclose(fp);
+out_free_mem:
+	free(status);
+	return -EINVAL;
+}
+
+/*
+ * Read the content of /proc/stat into the input buffer.
+ * Used by functions doing cpu idle detection
+ */
+int read_proc_stat(char *buffer, int size)
+{
+	int position = 0;
+	int retval;
+	int fd;
+
+	fd = open("/proc/stat", O_RDONLY);
+
+	if (fd < 0)
+		goto out_error;
+
+	do {
+		retval = read(fd, &buffer[position], size - position);
+		if (retval < 0)
+			goto out_close_fd;
+
+		position += retval;
+
+	} while (retval > 0 && position < size);
+
+	buffer[position-1] = '\0';
+
+	close(fd);
+
+	return position;
+
+out_close_fd:
+	close(fd);
+
+out_error:
+	return 0;
+}
